@@ -18,7 +18,10 @@
 //!
 //! `MemoryPool` keeps a vector of transactions seen by the miner.
 
+use std::sync::atomic::{AtomicUsize, Ordering};
+
 use crate::error::ConsensusError;
+use mpmc_map::MpmcMap;
 use snarkos_storage::Ledger;
 use snarkvm_algorithms::traits::LoadableMerkleParameters;
 use snarkvm_dpc::{BlockHeader, LedgerScheme, Storage, TransactionScheme, Transactions as DPCTransactions};
@@ -28,8 +31,6 @@ use snarkvm_utilities::{
     to_bytes,
 };
 
-use std::collections::HashMap;
-
 /// Stores a transaction and it's size in the memory pool.
 #[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
 pub struct Entry<T: TransactionScheme> {
@@ -39,18 +40,27 @@ pub struct Entry<T: TransactionScheme> {
 
 /// Stores transactions received by the server.
 /// Transaction entries will eventually be fetched by the miner and assembled into blocks.
-#[derive(Debug, Clone)]
-pub struct MemoryPool<T: TransactionScheme> {
+#[derive(Debug)]
+pub struct MemoryPool<T: TransactionScheme + Send + Sync + 'static> {
     /// The mapping of all unconfirmed transaction IDs to their corresponding transaction data.
-    pub transactions: HashMap<Vec<u8>, Entry<T>>,
+    pub transactions: MpmcMap<Vec<u8>, Entry<T>>,
     /// The total size in bytes of the current memory pool.
-    pub total_size_in_bytes: usize,
+    pub total_size_in_bytes: AtomicUsize,
+}
+
+impl<T: TransactionScheme + Send + Sync + 'static> Clone for MemoryPool<T> {
+    fn clone(&self) -> Self {
+        Self {
+            transactions: self.transactions.clone(),
+            total_size_in_bytes: AtomicUsize::new(self.total_size_in_bytes.load(Ordering::SeqCst)),
+        }
+    }
 }
 
 const BLOCK_HEADER_SIZE: usize = BlockHeader::size();
 const COINBASE_TRANSACTION_SIZE: usize = 1490; // TODO Find the value for actual coinbase transaction size
 
-impl<T: TransactionScheme> MemoryPool<T> {
+impl<T: TransactionScheme + Send + Sync + 'static> MemoryPool<T> {
     /// Initialize a new memory pool with no transactions
     #[inline]
     pub fn new() -> Self {
@@ -58,10 +68,10 @@ impl<T: TransactionScheme> MemoryPool<T> {
     }
 
     /// Load the memory pool from previously stored state in storage
-    pub fn from_storage<P: LoadableMerkleParameters, S: Storage>(
+    pub async fn from_storage<P: LoadableMerkleParameters, S: Storage>(
         storage: &Ledger<T, P, S>,
     ) -> Result<Self, ConsensusError> {
-        let mut memory_pool = Self::new();
+        let memory_pool = Self::new();
 
         if let Ok(Some(serialized_transactions)) = storage.get_memory_pool() {
             if let Ok(transaction_bytes) = DPCTransactions::<T>::read(&serialized_transactions[..]) {
@@ -71,7 +81,7 @@ impl<T: TransactionScheme> MemoryPool<T> {
                         transaction,
                         size_in_bytes: size,
                     };
-                    memory_pool.insert(storage, entry)?;
+                    memory_pool.insert(storage, entry).await?;
                 }
             }
         }
@@ -87,7 +97,7 @@ impl<T: TransactionScheme> MemoryPool<T> {
     ) -> Result<(), ConsensusError> {
         let mut transactions = DPCTransactions::<T>::new();
 
-        for (_transaction_id, entry) in self.transactions.iter() {
+        for (_transaction_id, entry) in self.transactions.inner().iter() {
             transactions.push(entry.transaction.clone())
         }
 
@@ -99,8 +109,8 @@ impl<T: TransactionScheme> MemoryPool<T> {
     }
 
     /// Adds entry to memory pool if valid in the current ledger.
-    pub fn insert<P: LoadableMerkleParameters, S: Storage>(
-        &mut self,
+    pub async fn insert<P: LoadableMerkleParameters, S: Storage>(
+        &self,
         storage: &Ledger<T, P, S>,
         entry: Entry<T>,
     ) -> Result<Option<Vec<u8>>, ConsensusError> {
@@ -117,9 +127,10 @@ impl<T: TransactionScheme> MemoryPool<T> {
 
         let mut holding_serial_numbers = vec![];
         let mut holding_commitments = vec![];
-        let mut holding_memos = Vec::with_capacity(self.transactions.len());
+        let mut holding_memos = Vec::with_capacity(self.transactions.inner().len());
 
-        for (_, tx) in self.transactions.iter() {
+        let txns = self.transactions.inner();
+        for (_, tx) in txns.iter() {
             holding_serial_numbers.extend(tx.transaction.old_serial_numbers());
             holding_commitments.extend(tx.transaction.new_commitments());
             holding_memos.push(tx.transaction.memorandum());
@@ -143,39 +154,44 @@ impl<T: TransactionScheme> MemoryPool<T> {
 
         let transaction_id = entry.transaction.transaction_id()?.to_vec();
 
-        self.total_size_in_bytes += entry.size_in_bytes;
-        self.transactions.insert(transaction_id.clone(), entry);
+        self.total_size_in_bytes
+            .fetch_add(entry.size_in_bytes, Ordering::SeqCst);
+        self.transactions.insert(transaction_id.clone(), entry).await;
 
         Ok(Some(transaction_id))
     }
 
     /// Cleanse the memory pool of outdated transactions.
     #[inline]
-    pub fn cleanse<P: LoadableMerkleParameters, S: Storage>(
-        &mut self,
+    pub async fn cleanse<P: LoadableMerkleParameters, S: Storage>(
+        &self,
         storage: &Ledger<T, P, S>,
     ) -> Result<(), ConsensusError> {
-        let mut new_memory_pool = Self::new();
+        let new_memory_pool = Self::new();
 
-        for (_, entry) in self.clone().transactions.iter() {
-            new_memory_pool.insert(&storage, entry.clone())?;
+        for (_, entry) in self.clone().transactions.inner().iter() {
+            new_memory_pool.insert(storage, entry.clone()).await?;
         }
 
-        self.total_size_in_bytes = new_memory_pool.total_size_in_bytes;
-        self.transactions = new_memory_pool.transactions;
+        self.total_size_in_bytes.store(
+            new_memory_pool.total_size_in_bytes.load(Ordering::SeqCst),
+            Ordering::SeqCst,
+        );
+        self.transactions.reset(new_memory_pool.transactions.inner_full());
 
         Ok(())
     }
 
     /// Removes transaction from memory pool or error.
     #[inline]
-    pub fn remove(&mut self, entry: &Entry<T>) -> Result<Option<Vec<u8>>, ConsensusError> {
+    pub async fn remove(&self, entry: &Entry<T>) -> Result<Option<Vec<u8>>, ConsensusError> {
         if self.contains(entry) {
-            self.total_size_in_bytes -= entry.size_in_bytes;
+            self.total_size_in_bytes
+                .fetch_sub(entry.size_in_bytes, Ordering::SeqCst);
 
             let transaction_id = entry.transaction.transaction_id()?.to_vec();
 
-            self.transactions.remove(&transaction_id);
+            self.transactions.remove(transaction_id.to_vec()).await;
 
             return Ok(Some(transaction_id));
         }
@@ -185,11 +201,13 @@ impl<T: TransactionScheme> MemoryPool<T> {
 
     /// Removes transaction from memory pool based on the transaction id.
     #[inline]
-    pub fn remove_by_hash(&mut self, transaction_id: &[u8]) -> Result<Option<Entry<T>>, ConsensusError> {
-        match self.transactions.clone().get(transaction_id) {
+    pub async fn remove_by_hash(&self, transaction_id: &[u8]) -> Result<Option<Entry<T>>, ConsensusError> {
+        match self.transactions.get(transaction_id) {
             Some(entry) => {
-                self.total_size_in_bytes -= entry.size_in_bytes;
-                self.transactions.remove(transaction_id);
+                self.total_size_in_bytes
+                    .fetch_sub(entry.size_in_bytes, Ordering::SeqCst);
+
+                self.transactions.remove(transaction_id.to_vec()).await;
 
                 Ok(Some(entry.clone()))
             }
@@ -218,7 +236,7 @@ impl<T: TransactionScheme> MemoryPool<T> {
         let mut transactions = DPCTransactions::new();
 
         // TODO Change naive transaction selection
-        for (_transaction_id, entry) in self.transactions.iter() {
+        for (_transaction_id, entry) in self.transactions.inner().iter() {
             if block_size + entry.size_in_bytes <= max_size {
                 if storage.transaction_conflicts(&entry.transaction) || transactions.conflicts(&entry.transaction) {
                     continue;
@@ -233,11 +251,11 @@ impl<T: TransactionScheme> MemoryPool<T> {
     }
 }
 
-impl<T: TransactionScheme> Default for MemoryPool<T> {
+impl<T: TransactionScheme + Send + Sync + 'static> Default for MemoryPool<T> {
     fn default() -> Self {
         Self {
-            total_size_in_bytes: 0,
-            transactions: HashMap::<Vec<u8>, Entry<T>>::new(),
+            total_size_in_bytes: AtomicUsize::new(0),
+            transactions: MpmcMap::<Vec<u8>, Entry<T>>::new(),
         }
     }
 }
@@ -250,11 +268,11 @@ mod tests {
 
     // MemoryPool tests use TRANSACTION_2 because memory pools shouldn't store coinbase transactions
 
-    #[test]
-    fn push() {
+    #[tokio::test]
+    async fn push() {
         let blockchain = FIXTURE_VK.ledger();
 
-        let mut mem_pool = MemoryPool::new();
+        let mem_pool = MemoryPool::new();
         let transaction = Tx::read(&TRANSACTION_2[..]).unwrap();
         let size = TRANSACTION_2.len();
 
@@ -263,9 +281,10 @@ mod tests {
                 size_in_bytes: size,
                 transaction: transaction.clone(),
             })
+            .await
             .unwrap();
 
-        assert_eq!(size, mem_pool.total_size_in_bytes);
+        assert_eq!(size, mem_pool.total_size_in_bytes.load(Ordering::SeqCst));
         assert_eq!(1, mem_pool.transactions.len());
 
         // Duplicate pushes don't do anything
@@ -275,17 +294,18 @@ mod tests {
                 size_in_bytes: size,
                 transaction,
             })
+            .await
             .unwrap();
 
-        assert_eq!(size, mem_pool.total_size_in_bytes);
+        assert_eq!(size, mem_pool.total_size_in_bytes.load(Ordering::SeqCst));
         assert_eq!(1, mem_pool.transactions.len());
     }
 
-    #[test]
-    fn remove_entry() {
+    #[tokio::test]
+    async fn remove_entry() {
         let blockchain = FIXTURE_VK.ledger();
 
-        let mut mem_pool = MemoryPool::new();
+        let mem_pool = MemoryPool::new();
         let transaction = Tx::read(&TRANSACTION_2[..]).unwrap();
         let size = TRANSACTION_2.len();
 
@@ -294,22 +314,22 @@ mod tests {
             transaction,
         };
 
-        mem_pool.insert(&blockchain, entry.clone()).unwrap();
+        mem_pool.insert(&blockchain, entry.clone()).await.unwrap();
 
         assert_eq!(1, mem_pool.transactions.len());
-        assert_eq!(size, mem_pool.total_size_in_bytes);
+        assert_eq!(size, mem_pool.total_size_in_bytes.load(Ordering::SeqCst));
 
-        mem_pool.remove(&entry).unwrap();
+        mem_pool.remove(&entry).await.unwrap();
 
         assert_eq!(0, mem_pool.transactions.len());
-        assert_eq!(0, mem_pool.total_size_in_bytes);
+        assert_eq!(0, mem_pool.total_size_in_bytes.load(Ordering::SeqCst));
     }
 
-    #[test]
-    fn remove_transaction_by_hash() {
+    #[tokio::test]
+    async fn remove_transaction_by_hash() {
         let blockchain = FIXTURE_VK.ledger();
 
-        let mut mem_pool = MemoryPool::new();
+        let mem_pool = MemoryPool::new();
         let transaction = Tx::read(&TRANSACTION_2[..]).unwrap();
         let size = TRANSACTION_2.len();
 
@@ -318,24 +338,26 @@ mod tests {
                 size_in_bytes: size,
                 transaction: transaction.clone(),
             })
+            .await
             .unwrap();
 
         assert_eq!(1, mem_pool.transactions.len());
-        assert_eq!(size, mem_pool.total_size_in_bytes);
+        assert_eq!(size, mem_pool.total_size_in_bytes.load(Ordering::SeqCst));
 
         mem_pool
             .remove_by_hash(&transaction.transaction_id().unwrap().to_vec())
+            .await
             .unwrap();
 
         assert_eq!(0, mem_pool.transactions.len());
-        assert_eq!(0, mem_pool.total_size_in_bytes);
+        assert_eq!(0, mem_pool.total_size_in_bytes.load(Ordering::SeqCst));
     }
 
-    #[test]
-    fn get_candidates() {
+    #[tokio::test]
+    async fn get_candidates() {
         let blockchain = FIXTURE_VK.ledger();
 
-        let mut mem_pool = MemoryPool::new();
+        let mem_pool = MemoryPool::new();
         let transaction = Tx::read(&TRANSACTION_2[..]).unwrap();
 
         let size = to_bytes![transaction].unwrap().len();
@@ -346,6 +368,7 @@ mod tests {
                 size_in_bytes: size,
                 transaction,
             })
+            .await
             .unwrap();
 
         let max_block_size = size + BLOCK_HEADER_SIZE + COINBASE_TRANSACTION_SIZE;
@@ -355,39 +378,44 @@ mod tests {
         assert!(candidates.contains(&expected_transaction));
     }
 
-    #[test]
-    fn store_memory_pool() {
+    #[tokio::test]
+    async fn store_memory_pool() {
         let blockchain = FIXTURE_VK.ledger();
 
-        let mut mem_pool = MemoryPool::new();
+        let mem_pool = MemoryPool::new();
         let transaction = Tx::read(&TRANSACTION_2[..]).unwrap();
         mem_pool
             .insert(&blockchain, Entry {
                 size_in_bytes: TRANSACTION_2.len(),
                 transaction,
             })
+            .await
             .unwrap();
 
         assert_eq!(1, mem_pool.transactions.len());
 
         mem_pool.store(&blockchain).unwrap();
 
-        let new_mem_pool = MemoryPool::from_storage(&blockchain).unwrap();
+        let new_mem_pool = MemoryPool::from_storage(&blockchain).await.unwrap();
 
-        assert_eq!(mem_pool.total_size_in_bytes, new_mem_pool.total_size_in_bytes);
+        assert_eq!(
+            mem_pool.total_size_in_bytes.load(Ordering::SeqCst),
+            new_mem_pool.total_size_in_bytes.load(Ordering::SeqCst)
+        );
     }
 
-    #[test]
-    fn cleanse_memory_pool() {
+    #[tokio::test]
+    async fn cleanse_memory_pool() {
         let blockchain = FIXTURE_VK.ledger();
 
-        let mut mem_pool = MemoryPool::new();
+        let mem_pool = MemoryPool::new();
         let transaction = Tx::read(&TRANSACTION_2[..]).unwrap();
         mem_pool
             .insert(&blockchain, Entry {
                 size_in_bytes: TRANSACTION_2.len(),
                 transaction,
             })
+            .await
             .unwrap();
 
         assert_eq!(1, mem_pool.transactions.len());
@@ -400,9 +428,9 @@ mod tests {
         blockchain.insert_and_commit(&block_1).unwrap();
         blockchain.insert_and_commit(&block_2).unwrap();
 
-        mem_pool.cleanse(&blockchain).unwrap();
+        mem_pool.cleanse(&blockchain).await.unwrap();
 
         assert_eq!(0, mem_pool.transactions.len());
-        assert_eq!(0, mem_pool.total_size_in_bytes);
+        assert_eq!(0, mem_pool.total_size_in_bytes.load(Ordering::SeqCst));
     }
 }
